/*
    Flitsim : A fully adaptable X-Windows based network simulator
    Copyright (C) 1993  Patrick Gaughan, Sudha Yalamanchili, Peter Flur
                        and Binh Dao


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*--------------------------------------------------------------------*/
/* Routing module.  Instead of implementing each in a different       */
/*     program, we just implement them as different routines.         */
/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/
/* Note on Deadlock Detection:					      */
/* Those routing protocol functions that begin with 'dd_' support     */
/* the Cycle and Deadlock Detection mechanisms.  The goal is to have  */
/* a routing function 'dd_<name>' for every touting function <name>   */
/* until the two are integrated, and Deadlock detection is performed  */
/* with a switch						      */
/*--------------------------------------------------------------------*/

static char router_rcsid[] = "$Id: route2.c,v 1.10 1994/06/07 14:58:05 dao Exp dao $";

#include <stdio.h>
#include "dat2.h"
#include "dd.h"
#include "irnet.h"

#define PROF     1
#define UNPROF   2

double drand48();
extern node *nodes;
extern plink *plinks;
extern vlane *vlinks;
extern msg msgs[MAXMESSAGE];
extern circuit circs[MAXCIRC];
extern int M, INIT;
extern int tmp;
extern int comm_mech;
extern int SLOW, SAVED_SLOW, CUMWAIT;
extern int XDISP;
extern int NODES;
extern int TOKENSPEED, HALFTOKENSPEED;

extern int dimensions, *width;
extern int protocol,virts, backtrack, vlog, node_virts, use_virts;
extern int select_fnc,RADIUS,ORDER,vary_virts,flip_dim;
abs(x) { if (x>0) return x; else return -x; }

/* newly added - for Deadlock Detection */
extern int CD;
extern int DD;
extern int injport, consport;

#ifdef PLINK
extern int plinknode;
extern int use_pl_inj;
extern int use_pl_con;
#endif

/* Hong - for timeout using overall waiting time */
extern int sim_time;

extern int IRDUATORESTR;

/*--------------------------------------------------------------------*/
/* reserve_vlink()  Reserve a vlink and update its activity counters  */
/*--------------------------------------------------------------------*/
reserve_vlink(vl,tkn)
     int vl,tkn;
{
  int circ_no;
  circ_no = msgs[tkn].circ_index;
  vlinks[vl].busy = circ_no+1;
  vlinks[vl].forward_block = 0;
  vlinks[vl].back_block = 0;
  /*  vlinks[vl].security_level = circs[circ_no].security_level;  */
  circs[circ_no].links_res++;
  circs[circ_no].path_length++;
  inc_setup(vl);
}

/*--------------------------------------------------------------------*/
/* route_to_node()                                                    */
/*--------------------------------------------------------------------*/
route_to_node(nd, tkn) {
  int i, vl,l;
  int index;
  int circ_no;
  int error;

  for (i=0;i<consport; i++) {
#ifdef PLINK
  for (l=0; l < use_pl_con; l++) {
#endif
#ifdef PLINK
    vl = nlink_addr(nd, i, OUT, 0, l);
#else
    vl = nlink_addr(nd, i, OUT, 0);
#endif
    if (vlinks[vl].busy == 0) {
      reserve_vlink(vl, UnPr(tkn));

      if(CD)
      {
		index = UnPr(tkn);
		circ_no = msgs[index].circ_index;

      		error = RemoveAllReqVCsFromCircuit(circ_no);
           	if (error)
		{
		  printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'route_to_node'\n", circ_no);
		  PrintAdjacencyList();
		}

	        error = AddVCToCircuit(circ_no, vl);
	        if (error)
		{
		  printf("ERROR: from 'AddVCToCircuit' - circ_no: %d, vl: %d in 'route_to_node'\n", circ_no, vl);
		  PrintAdjacencyList();
		} 
	      } /* CD*/

      return vl;
    }
  }
#ifdef PLINK
}
#endif
  return -1;
}

int s[16],d[16],n[16];
int j, k, sign, gamma, v;
int can_wait, min_c, vl_min;

/*----- DISHA Variables ----- */
int TOKEN = 0, seize_count = 0, seize_wait_time = 0;

extern	int	WAITTIME;

/*----- DISHA Variables ----- */

/*--------------------------------------------------------------------*/
/* route() Takes a probe token, a node and the link statuses and      */
/*         decides which virtual channel to route the probe over.     */
/*         return value of -1 indicates the probe blocks.             */
/*--------------------------------------------------------------------*/
route(nd,tkn,hist) 
     int nd,tkn,hist;
{

  int result;
  /*  calc_prof_dist(nd,tkn); */
  switch(protocol) {
#ifdef IRREGULAR
  case IRRE_STATIC:
    result = irre_static(nd,tkn);
    break;
  case IRRE_MIN_TFA:
    result = irre_min_tfa(nd,tkn);
    break;
  case IRRE_UP_DOWN:
    result = irre_up_down(nd,tkn);
    break;
  case IRRE_UP_DOWN_DUATO:
    result = irre_up_down_duato(nd,tkn);
    break;

  default:
    result = irre_min_tfa(nd,tkn);
  }
#endif
  if (result < -1) 
    return route_to_node(nd, tkn);
  else
    return result;
}




/*------------------------------------------------------------------------*/
/* irre_static()   Static Routing for Irregular Network                   */
/*------------------------------------------------------------------------*/

int irre_static(int nd,int tkn)
  {
      int i,vl,pl;
      int circ_no;
      int error;
      int DeadlockIndex;

      i = UnPr(tkn);
      circ_no = msgs[i].circ_index;
      
      if(nd==circs[circ_no].dest) return -nd-2;

      pl = StaticTableLookUp(nd, circs[circ_no].dest);

      for( i=0; i< use_virts ;i++)
	{
	  vl = link_addr(nd,pl,1,i,0);

/****** TESTING ***/
      
/*      
	printf("*** In 'irre_static' - NODE: %d,  MESSAGE: %d,  PL: %d,  VL: %d,  BUSY: %d\n", nd, circ_no, pl, vl,vlinks[vl].busy);
*/

/******************/

	  if(vlinks[vl].busy == 0)
	    {

	      reserve_vlink(vl,UnPr(tkn));

	      if(CD)
	      {
	        error = RemoveAllReqVCsFromCircuit(circ_no);
	        if (error)
		{
		  printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre_static'\n", circ_no);
		  PrintAdjacencyList();
		}
	        error = AddVCToCircuit(circ_no, vl);

	        if (error)
		{
		  printf("ERROR: from 'AddVCToCircuit' - circ_no: %d, vl: %d in 'irre_static'\n", circ_no, vl);
		  PrintAdjacencyList();
		} 
	      } /* CD*/
	      return vl;

	    }
	  else
	    {
	      if(CD)
	      {
	        error = AddReqVCToCircuit(circ_no, vl);
	        if (error)
	        {
		   printf("ERROR: from 'AddReqVCToCircuit' - circ_no: %d, vl: %d in 'irre_static'\n", circ_no, vl);
		   PrintAdjacencyList();
	        }
	      } /*CD*/
	    }
	}

      if(CD)
      {
	if (IsMessageInADeadlock(circ_no, &DeadlockIndex))
        {

	  /* Remove all ReqVCs and ... */
	  error = RemoveAllReqVCsFromCircuit(circ_no);

	  /* if there's an error, report it */
	  if (error)
	  {
	    printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre_static'\n", circ_no);
	    PrintAdjacencyList();
	  }

	  /* Deactivate the Deadlock ... since we broke it */
	  error = RetireDeadlock(DeadlockIndex);

	  if (error)
	  {
	    printf("ERROR: from 'RetireDeadlock' in 'irre'\n");
	  }

	  circs[circ_no].dest = nd;
	  return -nd-2;
        }
      }/*CD*/

      return -1;
  }



/*----------------------------------------------------------------------------*/
/* irre_min_tfa()  Minimal True Fully Adaptive Routing for Irregulat Network  */
/*----------------------------------------------------------------------------*/

int irre_min_tfa(int nd,int tkn)
  {
      int i,j, vl,pl;
      int circ_no;
      int error;
      int DeadlockIndex;
      TABLEENTRYPTR TableEntry;
      int NumOfPLs, randomPL;
      CONNLISTITEMPTR	CurrentConnection, StartingConnection;

      i = UnPr(tkn);
      circ_no = msgs[i].circ_index;
      
      if(nd==circs[circ_no].dest) return -nd-2;

      /* Look up in the TFAR routing table to get a list of eligible PLs ("connections") */
      TableEntry = (TABLEENTRYPTR) AdaptiveTableLookUp(nd, circs[circ_no].dest);

      /* The total number of Pls available for use */
      NumOfPLs = TableEntry->NumOfConnections;

      /* Randomly select the first PL to consider */
      randomPL = (int) (drand48() * NumOfPLs);
      CurrentConnection = TableEntry->FirstConnection;
      for(j=0 ; j < randomPL ; j++)
          CurrentConnection = CurrentConnection->NextConnection;
      StartingConnection = CurrentConnection;

      do 
      {
        pl = CurrentConnection->ConnectionID;
  
        for(i=0 ; i < use_virts ; i++)
  	{
	  vl = link_addr(nd,pl,1,i,0);
	  if(vlinks[vl].busy == 0)
	    {

	      reserve_vlink(vl,UnPr(tkn));

	      if(CD)
	      {
	        error = RemoveAllReqVCsFromCircuit(circ_no);
	        if (error)
		{
		  printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre'\n", circ_no);
		  PrintAdjacencyList();
		}
	        error = AddVCToCircuit(circ_no, vl);

	        if (error)
		{
		  printf("ERROR: from 'AddVCToCircuit' - circ_no: %d, vl: %d in 'irre'\n", circ_no, vl);
		  PrintAdjacencyList();
		} 
	      } /* CD*/
	      return vl;

	    }
	  else
	    {
	      if(CD)
	      {
	        error = AddReqVCToCircuit(circ_no, vl);
	        if (error)
	        {
		   printf("ERROR: from 'AddReqVCToCircuit' - circ_no: %d, vl: %d in 'irre'\n", circ_no, vl);
		   PrintAdjacencyList();
	        }
	      } /*CD*/
	    }
	}

        CurrentConnection = CurrentConnection->NextConnection;

      }
      while (CurrentConnection != StartingConnection);


      if(CD)
      {
	if (IsMessageInADeadlock(circ_no, &DeadlockIndex))
        {

	  /* Remove all ReqVCs and ... */
	  error = RemoveAllReqVCsFromCircuit(circ_no);

	  /* if there's an error, report it */
	  if (error)
	  {
	    printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre'\n", circ_no);
	    PrintAdjacencyList();
	  }

	  /* Deactivate the Deadlock ... since we broke it */
	  error = RetireDeadlock(DeadlockIndex);

	  if (error)
	  {
	    printf("ERROR: from 'RetireDeadlock' in 'irre'\n");
	  }

	  circs[circ_no].dest = nd;
	  return -nd-2;
        }
      }/*CD*/

      /* Timeout added by Hong */
      else if (WAITTIME > 0 && (sim_time - msgs[i].t_inject) > WAITTIME) {
	circs[circ_no].dest = nd;
	msgs[i].killed = 1;
	return -nd-2;
      }
      /***/

      return -1;
  }
	     




/*
*
*	irre_up_down
*
*	Given the current node and destination node, 
*	uses the Up/Down routing table to get the list
*	of eligible connections that can be used.
*	By starting at a randomly selected (eligible)
*	physical link, iterates through all of the 
*	virtual channels for all of the eligible physical
*	links until a free VC is found (and reserves it).
*	Returns either the Virtual Channel ID (vl) or 
*	-1 (if no free VC is found).
*
*	This is a deadlock avoidance algorithm. So no
*	deadlocks should form (thus no need for deadlock
*	detection and recovery).
*
*	The Cycle/Deadlock detection code is there in 
*	order to build a CWG --- to count cycles, etc.
*
*/

int irre_up_down(int nd,int tkn)
{
	int i,j, vl,pl;
	int error;
	int circ_no;
	TABLEENTRYPTR TableEntry;
	int NumOfPLs, randomPL;
	CONNLISTITEMPTR	CurrentConnection, StartingConnection;
	int CurrentDirection, NewDirection;
	int LinkEligible;
	int NextNode;

	i = UnPr(tkn);
	circ_no = msgs[i].circ_index;
      
	if(nd==circs[circ_no].dest) return -nd-2;

	/* Look up in the Up/Down routing table to get a list of eligible PLs ("connections") */
	TableEntry = (TABLEENTRYPTR) UpDownTableLookUp(nd, circs[circ_no].dest);

	/* Get the total number of PLs available for use */
	NumOfPLs = TableEntry->NumOfConnections;

	/* Randomly select the first PL to consider */
	randomPL = (int) (drand48() * NumOfPLs);
	CurrentConnection = TableEntry->FirstConnection;
	for(j=0 ; j < randomPL ; j++)
		CurrentConnection = CurrentConnection->NextConnection;
	
	StartingConnection = CurrentConnection;
	CurrentDirection = circs[circ_no].last_channel_type;

	do 
	{
		pl = CurrentConnection->ConnectionID;
		
		/* check for link eligibility */
		NextNode = GetDestNodeFromConnectionID(nd, CurrentConnection->ConnectionID);
		NewDirection = DetermineDirection(nd, NextNode);
		LinkEligible = DetermineLinkEligibility(CurrentDirection, NewDirection);
		
		if (!LinkEligible) /* skip non-eligible links */
		{
			CurrentConnection = CurrentConnection->NextConnection;
			continue;
		}
  
		for(i=0 ; i < use_virts ; i++)
		{
			vl = link_addr(nd,pl,1,i,0);
			if (vlinks[vl].busy == 0)
			{
				reserve_vlink(vl,UnPr(tkn));

				if(CD)
				{
					error = RemoveAllReqVCsFromCircuit(circ_no);

					if (error)
					{
						printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre_up_down'\n", circ_no);
						PrintAdjacencyList();
					}

					error = AddVCToCircuit(circ_no, vl);

	        			if (error)
					{
		  				printf("ERROR: from 'AddVCToCircuit' - circ_no: %d, vl: %d in 'irre_up_down'\n", circ_no, vl);
		  				PrintAdjacencyList();
					} 
	      			} /* CD*/

				/* Set to the channel type of channel just reserved  to 'Original' */
				circs[circ_no].last_channel_type = NewDirection;

	      			return vl;

			}
			else
			{
				if(CD)
				{
					error = AddReqVCToCircuit(circ_no, vl);

					if (error)
	        			{
		  				 printf("ERROR: from 'AddReqVCToCircuit' - circ_no: %d, vl: %d in 'irre_up_down'\n", circ_no, vl);
		   				PrintAdjacencyList();
	        			}
				} /*CD*/
			}
			
		}

		CurrentConnection = CurrentConnection->NextConnection;

	}
	while (CurrentConnection != StartingConnection);

	return -1;
}



/*
*
*	irre_up_down_duato
*
*	Given the current node and destination node, 
*	uses the Up/Down routing table to get the list
*	of eligible connections that can be used.
*	By starting at a randomly selected (eligible)
*	physical link, iterates through all of the 
*	virtual channels for all of the eligible physical
*	links until a free VC is found (and reserves it).
*	Returns either the Virtual Channel ID (vl) or 
*	-1 (if no free VC is found).
*
*	This is a deadlock avoidance algorithm. So no
*	deadlocks should form (thus no need for deadlock
*	detection and recovery).
*
*	The Cycle/Deadlock detection code is there in 
*	order to build a CWG --- to count cycles, etc.
*
*/

int irre_up_down_duato(int nd,int tkn)
{
	int i,j, vl,pl;
	int error;
	int circ_no;
	TABLEENTRYPTR TableEntry;
	int NumOfPLs, randomPL;
	CONNLISTITEMPTR	CurrentConnection, StartingConnection;
	int CurrentDirection, NewDirection;
	int LinkEligible;
	int NextNode;

	i = UnPr(tkn);
	circ_no = msgs[i].circ_index;
      
	if (nd==circs[circ_no].dest) return -nd-2;

	/* If this is the first channel (within the network) to be acquired (UNDEFINED) */
	/* OR the previous channel used was a "New" channel, then:			*/
	/* - try to use a new channel along the min. path (using min. TFAR)		*/
	if ((circs[circ_no].last_channel_type == UNDEFINED_CHANNEL) || (circs[circ_no].last_channel_type == ADAPTIVE_CHANNEL))
	{
		/* In the first iteration (i=0), attempts to use the "New"		*/
		/* channels (VC0) of all of the minimal paths (from TFAR table).	*/
		/* In the second iteration (i=1), attempts to use the "original"	*/
		/* channels (VC0) of all of the minimal paths (from TFAR table).	*/

      		/* Look up in the TFAR routing table to get a list of eligible PLs ("connections") */
      		TableEntry = (TABLEENTRYPTR) AdaptiveTableLookUp(nd, circs[circ_no].dest);

		/* Get the total number of PLs available for use */
		NumOfPLs = TableEntry->NumOfConnections;

		/* Randomly select the first PL to consider */
		randomPL = (int) (drand48() * NumOfPLs);
		CurrentConnection = TableEntry->FirstConnection;
		for(j=0 ; j < randomPL ; j++)
			CurrentConnection = CurrentConnection->NextConnection;
		StartingConnection = CurrentConnection;

		do 
		{
			pl = CurrentConnection->ConnectionID;
  
			/* 4th parameter = 0 implies 'new' channel VC */
			vl = link_addr(nd,pl,1,0,0);

			if (vlinks[vl].busy == 0)
			{
				reserve_vlink(vl,UnPr(tkn));

				if(CD)
				{
					error = RemoveAllReqVCsFromCircuit(circ_no);

					if (error)
					{
						printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre_up_down'\n", circ_no);
						PrintAdjacencyList();
					}
	
					error = AddVCToCircuit(circ_no, vl);

	        			if (error)
					{
		  				printf("ERROR: from 'AddVCToCircuit' - circ_no: %d, vl: %d in 'irre_up_down'\n", circ_no, vl);
		  				PrintAdjacencyList();
					} 
	      			} /* CD */

				/* Set to the channel type of channel just reserved : 'New' or 'Original' */
				circs[circ_no].last_channel_type = ADAPTIVE_CHANNEL;

	      			return vl;

			}
			else
			{
				if(CD)
				{
					error = AddReqVCToCircuit(circ_no, vl);

					if (error)
	        			{
		  			 	printf("ERROR: from 'AddReqVCToCircuit' - circ_no: %d, vl: %d in 'irre_up_down'\n", circ_no, vl);
		   				PrintAdjacencyList();
	        			}
				} /* CD */
			}

			CurrentConnection = CurrentConnection->NextConnection;

		}
		while (CurrentConnection != StartingConnection);
	}

	/* Otherwise, we try to use an "Original" channel			*/
	/* (unless there is a restriction on the use of an original channel as	*/
	/* the first channel acquired within the network) 			*/

	if ((! IRDUATORESTR) || (circs[circ_no].last_channel_type != UNDEFINED_CHANNEL))
	{

		TableEntry = (TABLEENTRYPTR) UpDownTableLookUp(nd, circs[circ_no].dest);
	
		/* Get the total number of PLs available for use */
		NumOfPLs = TableEntry->NumOfConnections;

		/* Randomly select the first PL to consider */
		randomPL = (int) (drand48() * NumOfPLs);
		CurrentConnection = TableEntry->FirstConnection;
		for(j=0 ; j < randomPL ; j++)
			CurrentConnection = CurrentConnection->NextConnection;
		StartingConnection = CurrentConnection;
		CurrentDirection = circs[circ_no].last_channel_type;

		do 
		{
			pl = CurrentConnection->ConnectionID;

			/* check for link eligibility */
			NextNode = GetDestNodeFromConnectionID(nd, CurrentConnection->ConnectionID);
			NewDirection = DetermineDirection(nd, NextNode);
			LinkEligible = DetermineLinkEligibility(CurrentDirection, NewDirection);

			if (!LinkEligible) /* skip non-eligible links */
			{
				CurrentConnection = CurrentConnection->NextConnection;
				continue;
			}

			/* 4th parameter = 1 implies 'original' channel VC */
			vl = link_addr(nd,pl,1,1,0);

			if (vlinks[vl].busy == 0)
			{
				reserve_vlink(vl,UnPr(tkn));

				if(CD)
				{
					error = RemoveAllReqVCsFromCircuit(circ_no);

					if (error)
					{
						printf("ERROR: from 'RemoveAllReqVCsFromCircuit' - circ_no: %d in 'irre_up_down'\n", circ_no);
						PrintAdjacencyList();
					}

					error = AddVCToCircuit(circ_no, vl);	

		        		if (error)
					{
			  			printf("ERROR: from 'AddVCToCircuit' - circ_no: %d, vl: %d in 'irre_up_down'\n", circ_no, vl);
			  			PrintAdjacencyList();
					} 
	 	     		} /* CD*/

				/* Set to the channel type of channel just reserved  to 'Original' */
				circs[circ_no].last_channel_type = NewDirection;

		      		return vl;

			}
			else
			{
				if(CD)
				{
					error = AddReqVCToCircuit(circ_no, vl);

					if (error)
		        		{
			  			printf("ERROR: from 'AddReqVCToCircuit' - circ_no: %d, vl: %d in 'irre_up_down'\n", circ_no, vl);
			   			PrintAdjacencyList();
	      		  		}
				} /*CD*/
			}

			CurrentConnection = CurrentConnection->NextConnection;

		}
		while (CurrentConnection != StartingConnection);
	}

	return -1;
}



